@@ -43,10 +43,17 @@
 #include <linux/uaccess.h>
 #include <linux/kallsyms.h>
 #include <linux/version.h>
+#include <linux/string.h>
+#include <linux/stacktrace.h>
 #include <asm/stacktrace.h>
 #include <asm/cacheflush.h>
+#include <generated/utsrelease.h>
 #include "kpatch.h"
 
+#ifndef UTS_UBUNTU_RELEASE_ABI
+#define UTS_UBUNTU_RELEASE_ABI 0
+#endif
+
 #if !defined(CONFIG_FUNCTION_TRACER) || \
 	!defined(CONFIG_HAVE_FENTRY) || \
 	!defined(CONFIG_MODULES) || \
@@ -60,22 +67,19 @@ static DEFINE_HASHTABLE(kpatch_func_hash, KPATCH_HASH_BITS);
 
 static DEFINE_SEMAPHORE(kpatch_mutex);
 
-LIST_HEAD(kpmod_list);
+static LIST_HEAD(kpmod_list);
 
 static int kpatch_num_patched;
 
-static struct kobject *kpatch_root_kobj;
-struct kobject *kpatch_patches_kobj;
-EXPORT_SYMBOL_GPL(kpatch_patches_kobj);
-
-struct kpatch_backtrace_args {
-	struct kpatch_module *kpmod;
-	int ret;
-};
+struct kobject *kpatch_root_kobj;
+EXPORT_SYMBOL_GPL(kpatch_root_kobj);
 
 struct kpatch_kallsyms_args {
+	const char *objname;
 	const char *name;
 	unsigned long addr;
+	unsigned long count;
+	unsigned long pos;
 };
 
 /* this is a double loop, use goto instead of break */
@@ -133,6 +137,13 @@ static atomic_t kpatch_state;
 static int (*kpatch_set_memory_rw)(unsigned long addr, int numpages);
 static int (*kpatch_set_memory_ro)(unsigned long addr, int numpages);
 
+#define MAX_STACK_TRACE_DEPTH   64
+static unsigned long stack_entries[MAX_STACK_TRACE_DEPTH];
+struct stack_trace trace = {
+	.max_entries	= ARRAY_SIZE(stack_entries),
+	.entries	= &stack_entries[0],
+};
+
 static inline void kpatch_state_idle(void)
 {
 	int state = atomic_read(&kpatch_state);
@@ -194,16 +205,12 @@ static inline int kpatch_compare_addresses(unsigned long stack_addr,
 	return 0;
 }
 
-static void kpatch_backtrace_address_verify(void *data, unsigned long address,
-					    int reliable)
+static int kpatch_backtrace_address_verify(struct kpatch_module *kpmod,
+					   unsigned long address)
 {
-	struct kpatch_backtrace_args *args = data;
-	struct kpatch_module *kpmod = args->kpmod;
 	struct kpatch_func *func;
 	int i;
-
-	if (args->ret)
-		return;
+	int ret;
 
 	/* check kpmod funcs */
 	do_for_each_linked_func(kpmod, func) {
@@ -227,55 +234,27 @@ static void kpatch_backtrace_address_verify(void *data, unsigned long address,
 			func_name = active_func->name;
 		}
 
-		args->ret = kpatch_compare_addresses(address, func_addr,
-						     func_size, func_name);
-		if (args->ret)
-			return;
+		ret = kpatch_compare_addresses(address, func_addr,
+					       func_size, func_name);
+		if (ret)
+			return ret;
 	} while_for_each_linked_func();
 
 	/* in the replace case, need to check the func hash as well */
 	hash_for_each_rcu(kpatch_func_hash, i, func, node) {
 		if (func->op == KPATCH_OP_UNPATCH && !func->force) {
-			args->ret = kpatch_compare_addresses(address,
-							     func->new_addr,
-							     func->new_size,
-							     func->name);
-			if (args->ret)
-				return;
+			ret = kpatch_compare_addresses(address,
+						       func->new_addr,
+						       func->new_size,
+						       func->name);
+			if (ret)
+				return ret;
 		}
 	}
+
+	return ret;
 }
 
-static int kpatch_backtrace_stack(void *data, char *name)
-{
-	return 0;
-}
-
-static const struct stacktrace_ops kpatch_backtrace_ops = {
-	.address	= kpatch_backtrace_address_verify,
-	.stack		= kpatch_backtrace_stack,
-	.walk_stack	= print_context_stack_bp,
-};
-
-static int kpatch_print_trace_stack(void *data, char *name)
-{
-	pr_cont(" <%s> ", name);
-	return 0;
-}
-
-static void kpatch_print_trace_address(void *data, unsigned long addr,
-				       int reliable)
-{
-	if (reliable)
-		pr_info("[<%p>] %pB\n", (void *)addr, (void *)addr);
-}
-
-static const struct stacktrace_ops kpatch_print_trace_ops = {
-	.stack		= kpatch_print_trace_stack,
-	.address	= kpatch_print_trace_address,
-	.walk_stack	= print_context_stack,
-};
-
 /*
  * Verify activeness safety, i.e. that none of the to-be-patched functions are
  * on the stack of any task.
@@ -285,26 +264,44 @@ static const struct stacktrace_ops kpatch_print_trace_ops = {
 static int kpatch_verify_activeness_safety(struct kpatch_module *kpmod)
 {
 	struct task_struct *g, *t;
+	int i;
 	int ret = 0;
 
-	struct kpatch_backtrace_args args = {
-		.kpmod = kpmod,
-		.ret = 0
-	};
-
 	/* Check the stacks of all tasks. */
 	do_each_thread(g, t) {
-		dump_trace(t, NULL, NULL, 0, &kpatch_backtrace_ops, &args);
-		if (args.ret) {
-			ret = args.ret;
-			pr_info("PID: %d Comm: %.20s\n", t->pid, t->comm);
-			dump_trace(t, NULL, (unsigned long *)t->thread.sp,
-				   0, &kpatch_print_trace_ops, NULL);
+
+		trace.nr_entries = 0;
+		save_stack_trace_tsk(t, &trace);
+		if (trace.nr_entries >= trace.max_entries) {
+			ret = -EBUSY;
+			pr_err("more than %u trace entries!\n",
+			       trace.max_entries);
 			goto out;
 		}
+
+                for (i = 0; i < trace.nr_entries; i++) {
+			if (trace.entries[i] == ULONG_MAX)
+				break;
+			ret = kpatch_backtrace_address_verify(kpmod,
+							      trace.entries[i]);
+			if (ret)
+				goto out;
+		}
+
 	} while_each_thread(g, t);
 
 out:
+	if (ret) {
+		pr_err("PID: %d Comm: %.20s\n", t->pid, t->comm);
+		for (i = 0; i < trace.nr_entries; i++) {
+			if (trace.entries[i] == ULONG_MAX)
+				break;
+			pr_err("  [<%pK>] %pB\n",
+			       (void *)trace.entries[i],
+			       (void *)trace.entries[i]);
+		}
+	}
+
 	return ret;
 }
 
@@ -449,9 +446,13 @@ done:
 	preempt_enable_notrace();
 }
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
+#define FTRACE_OPS_FL_IPMODIFY 0
+#endif
+
 static struct ftrace_ops kpatch_ftrace_ops __read_mostly = {
 	.func = kpatch_ftrace_handler,
-	.flags = FTRACE_OPS_FL_SAVE_REGS,
+	.flags = FTRACE_OPS_FL_SAVE_REGS | FTRACE_OPS_FL_IPMODIFY,
 };
 
 static int kpatch_ftrace_add_func(unsigned long ip)
@@ -512,58 +513,75 @@ static int kpatch_kallsyms_callback(void *data, const char *name,
 					 unsigned long addr)
 {
 	struct kpatch_kallsyms_args *args = data;
+	bool vmlinux = !strcmp(args->objname, "vmlinux");
 
-	if (args->addr == addr && !strcmp(args->name, name))
+	if ((mod && vmlinux) || (!mod && !vmlinux))
+		return 0;
+
+	if (strcmp(args->name, name))
+		return 0;
+
+	if (!vmlinux && strcmp(args->objname, mod->name))
+		return 0;
+
+	args->addr = addr;
+	args->count++;
+
+	/*
+	 * Finish the search when the symbol is found for the desired position
+	 * or the position is not defined for a non-unique symbol.
+	 */
+	if ((args->pos && (args->count == args->pos)) ||
+	    (!args->pos && (args->count > 1))) {
 		return 1;
-
-	return 0;
-}
-
-static int kpatch_verify_symbol_match(const char *name, unsigned long addr)
-{
-	int ret;
-
-	struct kpatch_kallsyms_args args = {
-		.name = name,
-		.addr = addr,
-	};
-
-	ret = kallsyms_on_each_symbol(kpatch_kallsyms_callback, &args);
-	if (!ret) {
-		pr_err("base kernel mismatch for symbol '%s'\n", name);
-		pr_err("expected address was 0x%016lx\n", addr);
-		return -EINVAL;
 	}
 
 	return 0;
 }
 
-static unsigned long kpatch_find_module_symbol(struct module *mod,
-					       const char *name)
+static int kpatch_find_object_symbol(const char *objname, const char *name,
+				     unsigned long sympos, unsigned long *addr)
 {
-	char buf[KSYM_SYMBOL_LEN];
+	struct kpatch_kallsyms_args args = {
+		.objname = objname,
+		.name = name,
+		.addr = 0,
+		.count = 0,
+		.pos = sympos,
+	};
 
-	/* check total string length for overrun */
-	if (strlen(mod->name) + strlen(name) + 1 >= KSYM_SYMBOL_LEN) {
-		pr_err("buffer overrun finding symbol '%s' in module '%s'\n",
-		       name, mod->name);
+	mutex_lock(&module_mutex);
+	kallsyms_on_each_symbol(kpatch_kallsyms_callback, &args);
+	mutex_unlock(&module_mutex);
+
+	/*
+	 * Ensure an address was found. If sympos is 0, ensure symbol is unique;
+	 * otherwise ensure the symbol position count matches sympos.
+	 */
+	if (args.addr == 0)
+		pr_err("symbol '%s' not found in symbol table\n", name);
+	else if (args.count > 1 && sympos == 0) {
+		pr_err("unresolvable ambiguity for symbol '%s' in object '%s'\n",
+		       name, objname);
+	} else if (sympos != args.count && sympos > 0) {
+		pr_err("symbol position %lu for symbol '%s' in object '%s' not found\n",
+		       sympos, name, objname);
+	} else {
+		*addr = args.addr;
 		return 0;
 	}
 
-	/* encode symbol name as "mod->name:name" */
-	strcpy(buf, mod->name);
-	strcat(buf, ":");
-	strcat(buf, name);
-
-	return kallsyms_lookup_name(buf);
+	*addr = 0;
+	return -EINVAL;
 }
 
 /*
  * External symbols are located outside the parent object (where the parent
  * object is either vmlinux or the kmod being patched).
  */
-static unsigned long kpatch_find_external_symbol(struct kpatch_module *kpmod,
-						 const char *name)
+static int kpatch_find_external_symbol(const char *objname, const char *name,
+				       unsigned long sympos, unsigned long *addr)
+
 {
 	const struct kernel_symbol *sym;
 
@@ -571,11 +589,13 @@ static unsigned long kpatch_find_external_symbol(struct kpatch_module *kpmod,
 	preempt_disable();
 	sym = find_symbol(name, NULL, NULL, true, true);
 	preempt_enable();
-	if (sym)
-		return sym->value;
+	if (sym) {
+		*addr = sym->value;
+		return 0;
+	}
 
 	/* otherwise check if it's in another .o within the patch module */
-	return kpatch_find_module_symbol(kpmod->mod, name);
+	return kpatch_find_object_symbol(objname, name, sympos, addr);
 }
 
 static int kpatch_write_relocations(struct kpatch_module *kpmod,
@@ -584,38 +604,31 @@ static int kpatch_write_relocations(struct kpatch_module *kpmod,
 	int ret, size, readonly = 0, numpages;
 	struct kpatch_dynrela *dynrela;
 	u64 loc, val;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0)
-       unsigned long core = (unsigned long)kpmod->mod->core_layout.base;
-       unsigned long core_size = kpmod->mod->core_layout.size;
+#if (( LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0) ) || \
+     ( LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0) && \
+      UTS_UBUNTU_RELEASE_ABI >= 7 ) \
+    )
+	unsigned long core = (unsigned long)kpmod->mod->core_layout.base;
+	unsigned long core_size = kpmod->mod->core_layout.size;
 #else
-       unsigned long core = (unsigned long)kpmod->mod->module_core;
-       unsigned long core_size = kpmod->mod->core_size;
+	unsigned long core = (unsigned long)kpmod->mod->module_core;
+	unsigned long core_size = kpmod->mod->core_size;
 #endif
-	unsigned long src;
 
 	list_for_each_entry(dynrela, &object->dynrelas, list) {
-		if (!strcmp(object->name, "vmlinux")) {
-			ret = kpatch_verify_symbol_match(dynrela->name,
-							 dynrela->src);
-			if (ret)
-				return ret;
-		} else {
-			/* module, dynrela->src needs to be discovered */
-
-			if (dynrela->external)
-				src = kpatch_find_external_symbol(kpmod,
-								  dynrela->name);
-			else
-				src = kpatch_find_module_symbol(object->mod,
-								dynrela->name);
-
-			if (!src) {
-				pr_err("unable to find symbol '%s'\n",
-				       dynrela->name);
-				return -EINVAL;
-			}
-
-			dynrela->src = src;
+		if (dynrela->external)
+			ret = kpatch_find_external_symbol(kpmod->mod->name,
+							  dynrela->name,
+							  dynrela->sympos,
+							  &dynrela->src);
+		else
+			ret = kpatch_find_object_symbol(object->name,
+							dynrela->name,
+							dynrela->sympos,
+							&dynrela->src);
+		if (ret) {
+			pr_err("unable to find symbol '%s'\n", dynrela->name);
+			return ret;
 		}
 
 		switch (dynrela->type) {
@@ -650,13 +663,28 @@ static int kpatch_write_relocations(struct kpatch_module *kpmod,
 			return -EINVAL;
 		}
 
-#ifdef CONFIG_DEBUG_SET_MODULE_RONX
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0)
-               if (loc < core + kpmod->mod->core_layout.ro_size)
+		/*
+		 * Skip it if the instruction to be relocated has been
+		 * changed already (paravirt or alternatives may do this).
+		 */
+		if (memchr_inv((void *)loc, 0, size)) {
+			pr_notice("Skipped dynrela for %s (0x%lx <- 0x%lx): the instruction has been changed already.\n",
+				  dynrela->name, dynrela->dest, dynrela->src);
+			pr_notice_once(
+"This is not necessarily a bug but it may indicate in some cases "
+"that the binary patch does not handle paravirt operations, alternatives or the like properly.\n");
+			continue;
+		}
+
+#if defined(CONFIG_DEBUG_SET_MODULE_RONX) || defined(CONFIG_ARCH_HAS_SET_MEMORY)
+#if (( LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0) ) || \
+     ( LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0) && \
+      UTS_UBUNTU_RELEASE_ABI >= 7 ) \
+    )
+               readonly = (loc < core + kpmod->mod->core_layout.ro_size);
 #else
-               if (loc < core + kpmod->mod->core_ro_size)
+               readonly = (loc < core + kpmod->mod->core_ro_size);
 #endif
-			readonly = 1;
 #endif
 
 		numpages = (PAGE_SIZE - (loc & ~PAGE_MASK) >= size) ? 1 : 2;
@@ -741,25 +769,14 @@ static int kpatch_link_object(struct kpatch_module *kpmod,
 
 	list_for_each_entry(func, &object->funcs, list) {
 
-		/* calculate actual old location */
-		if (vmlinux) {
-			ret = kpatch_verify_symbol_match(func->name,
-							 func->old_addr);
-			if (ret) {
-				func_err = func;
-				goto err_ftrace;
-			}
-		} else {
-			unsigned long old_addr;
-			old_addr = kpatch_find_module_symbol(mod, func->name);
-			if (!old_addr) {
-				pr_err("unable to find symbol '%s' in module '%s\n",
-				       func->name, mod->name);
-				func_err = func;
-				ret = -EINVAL;
-				goto err_ftrace;
-			}
-			func->old_addr = old_addr;
+		/* lookup the old location */
+		ret = kpatch_find_object_symbol(object->name,
+						func->name,
+						func->sympos,
+						&func->old_addr);
+		if (ret) {
+			func_err = func;
+			goto err_ftrace;
 		}
 
 		/* add to ftrace filter and register handler if needed */
@@ -957,9 +974,30 @@ int kpatch_register(struct kpatch_module *kpmod, bool replace)
 		func->op = KPATCH_OP_NONE;
 	} while_for_each_linked_func();
 
-	/* TODO: need TAINT_KPATCH */
+/* HAS_MODULE_TAINT - upstream 2992ef29ae01 "livepatch/module: make TAINT_LIVEPATCH module-specific" */
+/* HAS_MODULE_TAINT_LONG - upstream 7fd8329ba502 "taint/module: Clean up global and module taint flags handling" */
+#ifdef RHEL_RELEASE_CODE
+# if RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(7, 4)
+#  define HAS_MODULE_TAINT
+# endif
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)
+# define HAS_MODULE_TAINT_LONG
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
+# define HAS_MODULE_TAINT
+#endif
+
+#ifdef TAINT_LIVEPATCH
+	pr_notice_once("tainting kernel with TAINT_LIVEPATCH\n");
+	add_taint(TAINT_LIVEPATCH, LOCKDEP_STILL_OK);
+# ifdef HAS_MODULE_TAINT_LONG
+	set_bit(TAINT_LIVEPATCH, &kpmod->mod->taints);
+# elif defined(HAS_MODULE_TAINT)
+	kpmod->mod->taints |= (1 << TAINT_LIVEPATCH);
+# endif
+#else
 	pr_notice_once("tainting kernel with TAINT_USER\n");
 	add_taint(TAINT_USER, LOCKDEP_STILL_OK);
+#endif
 
 	pr_notice("loaded patch module '%s'\n", kpmod->mod->name);
 
@@ -1087,21 +1125,12 @@ static int kpatch_init(void)
 	if (!kpatch_root_kobj)
 		return -ENOMEM;
 
-	kpatch_patches_kobj = kobject_create_and_add("patches",
-						     kpatch_root_kobj);
-	if (!kpatch_patches_kobj) {
-		ret = -ENOMEM;
-		goto err_root_kobj;
-	}
-
 	ret = register_module_notifier(&kpatch_module_nb);
 	if (ret)
-		goto err_patches_kobj;
+		goto err_root_kobj;
 
 	return 0;
 
-err_patches_kobj:
-	kobject_put(kpatch_patches_kobj);
 err_root_kobj:
 	kobject_put(kpatch_root_kobj);
 	return ret;
@@ -1113,7 +1142,6 @@ static void kpatch_exit(void)
 
 	WARN_ON(kpatch_num_patched != 0);
 	WARN_ON(unregister_module_notifier(&kpatch_module_nb));
-	kobject_put(kpatch_patches_kobj);
 	kobject_put(kpatch_root_kobj);
 }
 
